@@ -109,7 +109,7 @@ typedef struct p80211ioctl_req
 	UINT32	magic				__WLAN_ATTRIB_PACK__;
 	UINT16	len				__WLAN_ATTRIB_PACK__;
 	UINT32	result				__WLAN_ATTRIB_PACK__;
-} p80211ioctl_req_t;
+} __WLAN_ATTRIB_PACK__ p80211ioctl_req_t;
 
 
 /*================================================================*/
