@@ -87,6 +87,9 @@
 						   to a textual name */
 #define P80211_TYPE_UNKDATA		6	/* Data item containing an
 						   unknown data type */
+#define P80211_TYPE_INTARRAY		7	/* Array of 32-bit integers. */
+#define P80211_TYPE_BITARRAY		8	/* Array of bits. */
+#define P80211_TYPE_MACARRAY		9	/* Array of MAC addresses. */
 
 /*----------------------------------------------------------------*/
 /* The following constants are indexes into the Mib Category List */
@@ -116,8 +119,11 @@
 /* p80211 DID field codes that represent access type and */
 /* is_table status. */
 
-#define P80211DID_ACCESS_READONLY	0
-#define P80211DID_ACCESS_READWRITE	1
+#define P80211DID_ACCESS_READ		0x10000000
+#define P80211DID_ACCESS_WRITE		0x08000000
+#define P80211DID_WRITEONLY		0x00000001
+#define P80211DID_READONLY		0x00000002
+#define P80211DID_READWRITE		0x00000003
 #define P80211DID_ISTABLE_FALSE		0
 #define P80211DID_ISTABLE_TRUE		1
 
@@ -132,6 +138,9 @@
 
 #define P80211ENUM_truth_false			0
 #define P80211ENUM_truth_true			1
+#define P80211ENUM_ifstate_disable		0
+#define P80211ENUM_ifstate_fwload		1
+#define P80211ENUM_ifstate_enable		2
 #define P80211ENUM_powermgmt_active		1
 #define P80211ENUM_powermgmt_powersave		2
 #define P80211ENUM_bsstype_infrastructure	1
@@ -160,6 +169,7 @@
 #define P80211ENUM_diversity_dynamic		3
 #define P80211ENUM_scantype_active		1
 #define P80211ENUM_scantype_passive		2
+#define P80211ENUM_scantype_both		3
 #define P80211ENUM_resultcode_success		1
 #define P80211ENUM_resultcode_invalid_parameters	2
 #define P80211ENUM_resultcode_not_supported	3
@@ -171,6 +181,7 @@
 #define P80211ENUM_resultcode_invalid_mibattribute	9
 #define P80211ENUM_resultcode_cant_set_readonly_mib	10
 #define P80211ENUM_resultcode_implementation_failure	11
+#define P80211ENUM_resultcode_cant_get_writeonly_mib	12
 #define P80211ENUM_reason_unspec_reason		1
 #define P80211ENUM_reason_auth_not_valid	2
 #define P80211ENUM_reason_deauth_lv_ss		3
@@ -191,6 +202,9 @@
 #define P80211ENUM_status_auth_timeout		16
 #define P80211ENUM_status_ap_full		17
 #define P80211ENUM_status_unsup_rate		18
+#define P80211ENUM_status_unsup_shortpreamble	19
+#define P80211ENUM_status_unsup_pbcc		20
+#define P80211ENUM_status_unsup_agility		21
 #define P80211ENUM_msgitem_status_data_ok		0
 #define P80211ENUM_msgitem_status_no_value		1
 #define P80211ENUM_msgitem_status_invalid_itemname	2
@@ -208,6 +222,10 @@
 #define P80211ENUM_msgitem_status_invalid_did		14
 #define P80211ENUM_msgitem_status_missing_print_func	15
 
+#define P80211ENUM_p2preamble_long               0
+#define P80211ENUM_p2preamble_short              2
+#define P80211ENUM_p2preamble_mixed              3
+
 /*----------------------------------------------------------------*/
 /* p80211 max length constants for the different pascal strings. */
 
@@ -215,11 +233,15 @@
 #define MAXLEN_PSTR14		(14)	/* pascal array of 14 bytes */
 #define MAXLEN_PSTR32		(32)	/* pascal array of 32 bytes */
 #define MAXLEN_PSTR255		(255)	/* pascal array of 255 bytes */
-#define MAXLEN_MIBATTRIBUTE	(264)	/* maximum mibattribute */
+#define MAXLEN_MIBATTRIBUTE	(392)	/* maximum mibattribute */
 					/* where the size of the DATA itself */
 					/* is a DID-LEN-DATA triple */
-					/* with a max size of 4+4+256 */
+					/* with a max size of 4+4+384 */
 
+#define P80211_SET_INT(item, value) do { \
+	(item).data   = (value); \
+	(item).status = P80211ENUM_msgitem_status_data_ok; \
+	} while(0)
 /*----------------------------------------------------------------*/
 /* string constants */
 
@@ -248,7 +270,7 @@
 /*  whether the metadata item is for a request command, confirm       */
 /*  command or both.                                                  */
 /*--------------------------------------------------------------------*/
-/* MSM: Does these belong in p80211meta.h?  I'm not sure */
+/* MSM: Do these belong in p80211meta.h?  I'm not sure */
 
 #define P80211ITEM_SETFLAGS(q, r, c)	( q | r | c )
 
@@ -261,9 +283,19 @@
 
 #define MKENUMNAME(name) p80211enum_ ## name
 
-/*----------------------------------------------------------------*/
-/* The following constants and macros are used to construct and */
-/* deconstruct the Data ID codes. */
+/*----------------------------------------------------------------
+* The following constants and macros are used to construct and
+* deconstruct the Data ID codes.  The coding is as follows:
+*
+*     ...rwtnnnnnnnniiiiiiggggggssssss      s - Section
+*                                           g - Group
+*                                           i - Item
+*                                           n - Index
+*                                           t - Table flag
+*                                           w - Write flag
+*                                           r - Read flag
+*                                           . - Unused
+*/
 
 #define P80211DID_INVALID		0xffffffffUL
 #define P80211DID_VALID			0x00000000UL
@@ -273,16 +305,14 @@
 #define P80211DID_LSB_ITEM		(12)
 #define P80211DID_LSB_INDEX		(18)
 #define P80211DID_LSB_ISTABLE		(26)
-#define P80211DID_LSB_ACCESS		(27)
-#define P80211DID_LSB_TYPE		(28)
+#define P80211DID_LSB_ACCESS 		(27)
 
 #define P80211DID_MASK_SECTION		(0x0000003fUL)
 #define P80211DID_MASK_GROUP		(0x0000003fUL)
 #define P80211DID_MASK_ITEM		(0x0000003fUL)
 #define P80211DID_MASK_INDEX		(0x000000ffUL)
 #define P80211DID_MASK_ISTABLE		(0x00000001UL)
-#define P80211DID_MASK_ACCESS		(0x00000001UL)
-#define P80211DID_MASK_TYPE		(0x00000007UL)
+#define P80211DID_MASK_ACCESS 		(0x00000003UL)
 
 
 #define P80211DID_MK(a,m,l)	((((UINT32)(a)) & (m)) << (l))
@@ -302,21 +332,14 @@
 #define P80211DID_MKISTABLE(a)	P80211DID_MK(a, \
 					P80211DID_MASK_ISTABLE, \
 					P80211DID_LSB_ISTABLE )
-#define P80211DID_MKACCESS(a)	P80211DID_MK(a, \
-					P80211DID_MASK_ACCESS, \
-					P80211DID_LSB_ACCESS )
-#define P80211DID_MKTYPE(a)	P80211DID_MK(a, \
-					P80211DID_MASK_TYPE, \
-					P80211DID_LSB_TYPE )
 
 
-#define P80211DID_MKID(s,g,i,n,t,a,d)	(P80211DID_MKSECTION(s) | \
+#define P80211DID_MKID(s,g,i,n,t,a)	(P80211DID_MKSECTION(s) | \
 						P80211DID_MKGROUP(g) | \
 				 		P80211DID_MKITEM(i) | \
 				 		P80211DID_MKINDEX(n) | \
 						P80211DID_MKISTABLE(t) | \
-						P80211DID_MKACCESS(a) | \
-						P80211DID_MKTYPE(d) )
+						(a) )
 
 
 #define P80211DID_GET(a,m,l)	((((UINT32)(a)) >> (l)) & (m))
@@ -339,9 +362,6 @@
 #define P80211DID_ACCESS(a)	P80211DID_GET(a, \
 					P80211DID_MASK_ACCESS, \
 					P80211DID_LSB_ACCESS)
-#define P80211DID_TYPE(a)	P80211DID_GET(a, \
-					P80211DID_MASK_TYPE, \
-					P80211DID_LSB_TYPE)
 
 /*================================================================*/
 /* Types */
@@ -371,7 +391,7 @@ __WLAN_PRAGMA_PACK1__
 typedef struct p80211pstr
 {
 	UINT8		len								__WLAN_ATTRIB_PACK__;			
-} p80211pstr_t;
+} __WLAN_ATTRIB_PACK__ p80211pstr_t;
 __WLAN_PRAGMA_PACKDFLT__
 
 __WLAN_PRAGMA_PACK1__
@@ -379,7 +399,7 @@ typedef struct p80211pstrd
 {
 	UINT8		len			__WLAN_ATTRIB_PACK__;			
 	UINT8		data[0]			__WLAN_ATTRIB_PACK__;
-} p80211pstrd_t;
+} __WLAN_ATTRIB_PACK__ p80211pstrd_t;
 __WLAN_PRAGMA_PACKDFLT__
 
 /* Maximum pascal string */
@@ -388,7 +408,7 @@ typedef struct p80211pstr255
 {
 	UINT8		len			__WLAN_ATTRIB_PACK__;
 	UINT8		data[MAXLEN_PSTR255]	__WLAN_ATTRIB_PACK__;
-} p80211pstr255_t;
+} __WLAN_ATTRIB_PACK__ p80211pstr255_t;
 __WLAN_PRAGMA_PACKDFLT__
 
 /* pascal string for macaddress and bssid */
@@ -397,7 +417,7 @@ typedef struct p80211pstr6
 {
 	UINT8		len			__WLAN_ATTRIB_PACK__;
 	UINT8		data[MAXLEN_PSTR6]	__WLAN_ATTRIB_PACK__;
-} p80211pstr6_t;
+} __WLAN_ATTRIB_PACK__ p80211pstr6_t;
 __WLAN_PRAGMA_PACKDFLT__
 
 /* pascal string for channel list */
@@ -406,7 +426,7 @@ typedef struct p80211pstr14
 {
 	UINT8		len			__WLAN_ATTRIB_PACK__;
 	UINT8		data[MAXLEN_PSTR14]	__WLAN_ATTRIB_PACK__;
-} p80211pstr14_t;
+} __WLAN_ATTRIB_PACK__ p80211pstr14_t;
 __WLAN_PRAGMA_PACKDFLT__
 
 /* pascal string for ssid */
@@ -415,7 +435,16 @@ typedef struct p80211pstr32
 {
 	UINT8		len			__WLAN_ATTRIB_PACK__;
 	UINT8		data[MAXLEN_PSTR32]	__WLAN_ATTRIB_PACK__;
-} p80211pstr32_t;
+} __WLAN_ATTRIB_PACK__ p80211pstr32_t;
+__WLAN_PRAGMA_PACKDFLT__
+
+/* MAC address array */
+__WLAN_PRAGMA_PACK1__
+typedef struct p80211macarray
+{
+	UINT32		cnt			__WLAN_ATTRIB_PACK__;
+	UINT8		data[1][MAXLEN_PSTR6]	__WLAN_ATTRIB_PACK__;
+} __WLAN_ATTRIB_PACK__ p80211macarray_t;
 __WLAN_PRAGMA_PACKDFLT__
 
 /* prototype template */
@@ -425,7 +454,7 @@ typedef struct p80211item
 	UINT32		did			__WLAN_ATTRIB_PACK__;
 	UINT16		status			__WLAN_ATTRIB_PACK__;
 	UINT16		len			__WLAN_ATTRIB_PACK__;
-} p80211item_t;
+} __WLAN_ATTRIB_PACK__ p80211item_t;
 __WLAN_PRAGMA_PACKDFLT__
 
 /* prototype template w/ data item */
@@ -436,7 +465,7 @@ typedef struct p80211itemd
 	UINT16		status			__WLAN_ATTRIB_PACK__;
 	UINT16		len			__WLAN_ATTRIB_PACK__;
 	UINT8		data[0]			__WLAN_ATTRIB_PACK__;
-} p80211itemd_t;
+} __WLAN_ATTRIB_PACK__ p80211itemd_t;
 __WLAN_PRAGMA_PACKDFLT__
 
 /* message data item for INT, BOUNDEDINT, ENUMINT */
@@ -447,7 +476,7 @@ typedef struct p80211item_uint32
 	UINT16		status			__WLAN_ATTRIB_PACK__;
 	UINT16		len			__WLAN_ATTRIB_PACK__;
 	UINT32		data			__WLAN_ATTRIB_PACK__;
-} p80211item_uint32_t;
+} __WLAN_ATTRIB_PACK__ p80211item_uint32_t;
 __WLAN_PRAGMA_PACKDFLT__
 
 /* message data item for OCTETSTR, DISPLAYSTR */
@@ -458,7 +487,7 @@ typedef struct p80211item_pstr6
 	UINT16		status			__WLAN_ATTRIB_PACK__;
 	UINT16		len			__WLAN_ATTRIB_PACK__;
 	p80211pstr6_t	data			__WLAN_ATTRIB_PACK__;
-}p80211item_pstr6_t;
+} __WLAN_ATTRIB_PACK__ p80211item_pstr6_t;
 __WLAN_PRAGMA_PACKDFLT__
 
 /* message data item for OCTETSTR, DISPLAYSTR */
@@ -469,7 +498,7 @@ typedef struct p80211item_pstr14
 	UINT16			status		__WLAN_ATTRIB_PACK__;
 	UINT16			len		__WLAN_ATTRIB_PACK__;
 	p80211pstr14_t		data		__WLAN_ATTRIB_PACK__;
-}p80211item_pstr14_t;
+} __WLAN_ATTRIB_PACK__ p80211item_pstr14_t;
 __WLAN_PRAGMA_PACKDFLT__
 
 /* message data item for OCTETSTR, DISPLAYSTR */
@@ -480,7 +509,7 @@ typedef struct p80211item_pstr32
 	UINT16			status		__WLAN_ATTRIB_PACK__;
 	UINT16			len		__WLAN_ATTRIB_PACK__;
 	p80211pstr32_t		data		__WLAN_ATTRIB_PACK__;
-}p80211item_pstr32_t;
+} __WLAN_ATTRIB_PACK__ p80211item_pstr32_t;
 __WLAN_PRAGMA_PACKDFLT__
 
 /* message data item for OCTETSTR, DISPLAYSTR */
@@ -491,17 +520,17 @@ typedef struct p80211item_pstr255
 	UINT16			status		__WLAN_ATTRIB_PACK__;
 	UINT16			len		__WLAN_ATTRIB_PACK__;
 	p80211pstr255_t		data		__WLAN_ATTRIB_PACK__;
-}p80211item_pstr255_t;
+} __WLAN_ATTRIB_PACK__ p80211item_pstr255_t;
 __WLAN_PRAGMA_PACKDFLT__
 
-/* message data item for UNK 264, namely mib items */
-typedef struct  p80211item_unk264
+/* message data item for UNK 392, namely mib items */
+typedef struct  p80211item_unk392
 {
 	UINT32		did				__WLAN_ATTRIB_PACK__;
 	UINT16		status				__WLAN_ATTRIB_PACK__;
 	UINT16		len				__WLAN_ATTRIB_PACK__;
 	UINT8		data[MAXLEN_MIBATTRIBUTE]	__WLAN_ATTRIB_PACK__;
-} p80211item_unk264_t;
+} __WLAN_ATTRIB_PACK__ p80211item_unk392_t;
 __WLAN_PRAGMA_PACK1__
 
 /* message data item for UNK 1025, namely p2 pdas */
@@ -511,7 +540,7 @@ typedef struct  p80211item_unk1024
 	UINT16		status				__WLAN_ATTRIB_PACK__;
 	UINT16		len				__WLAN_ATTRIB_PACK__;
 	UINT8		data[1024]			__WLAN_ATTRIB_PACK__;
-} p80211item_unk1024_t;
+}  __WLAN_ATTRIB_PACK__ p80211item_unk1024_t;
 __WLAN_PRAGMA_PACK1__
 
 /* message data item for UNK 4096, namely p2 download chunks */
@@ -521,7 +550,7 @@ typedef struct  p80211item_unk4096
 	UINT16		status				__WLAN_ATTRIB_PACK__;
 	UINT16		len				__WLAN_ATTRIB_PACK__;
 	UINT8		data[4096]			__WLAN_ATTRIB_PACK__;
-} p80211item_unk4096_t;
+}  __WLAN_ATTRIB_PACK__ p80211item_unk4096_t;
 __WLAN_PRAGMA_PACK1__
 
 struct catlistitem;
@@ -545,6 +574,7 @@ typedef UINT32 (*p80211_valid_t)( struct catlistitem *, UINT32 did, UINT8* itemb
 /*  for all enumerations  */
 
 extern p80211enum_t MKENUMNAME(truth); 
+extern p80211enum_t MKENUMNAME(ifstate); 
 extern p80211enum_t MKENUMNAME(powermgmt);
 extern p80211enum_t MKENUMNAME(bsstype);
 extern p80211enum_t MKENUMNAME(authalg);
@@ -559,6 +589,7 @@ extern p80211enum_t MKENUMNAME(reason);
 extern p80211enum_t MKENUMNAME(status);
 extern p80211enum_t MKENUMNAME(msgcode);
 extern p80211enum_t MKENUMNAME(msgitem_status);
+extern p80211enum_t MKENUMNAME(p2preamble); 
 
 /*================================================================*/
 /* Function Declarations */
@@ -631,6 +662,34 @@ void p80211_fromtext_enumint( struct catlistitem *metalist, UINT32 did, UINT8 *i
 /* function that checks validity of an enum's binary value */
 UINT32 p80211_isvalid_enumint( struct catlistitem *metalist, UINT32 did, UINT8 *itembuf );
 
+/*-- INTARRAY --------------------------------------------------------*/
+/* UINT32[] => %d,%d,%d,... */
+void p80211_totext_intarray( struct catlistitem *metalist, UINT32 did, UINT8 *itembuf, char *textbuf );
+
+/* %d,%d,%d,... ==> UINT32[] */
+void p80211_fromtext_intarray( struct catlistitem *metalist, UINT32 did, UINT8 *itembuf, char *textbuf );
+
+/* function that checks validity of an integer array's value */
+UINT32 p80211_isvalid_intarray( struct catlistitem *metalist, UINT32 did, UINT8 *itembuf );
+
+/*-- BITARRAY --------------------------------------------------------*/
+/* UINT32 ==> %d,%d,%d,... */
+void p80211_totext_bitarray( struct catlistitem *metalist, UINT32 did, UINT8 *itembuf, char *textbuf );
+
+/* %d,%d,%d,... ==> UINT32 */
+void p80211_fromtext_bitarray( struct catlistitem *metalist, UINT32 did, UINT8 *itembuf, char *textbuf );
+
+/* function that checks validity of a bit array's value */
+UINT32 p80211_isvalid_bitarray( struct catlistitem *metalist, UINT32 did, UINT8 *itembuf );
+
+/*-- MACARRAY --------------------------------------------------------*/
+void p80211_totext_macarray( struct catlistitem *metalist, UINT32 did, UINT8 *itembuf, char *textbuf );
+
+void p80211_fromtext_macarray( struct catlistitem *metalist, UINT32 did, UINT8 *itembuf, char *textbuf );
+
+/* function that checks validity of a MAC address array's value */
+UINT32 p80211_isvalid_macarray( struct catlistitem *metalist, UINT32 did, UINT8 *itembuf );
+
 /*-- MIBATTRIUBTE ------------------------------------------------------*/
 /* <mibvalue> ==> <textual representation identified in MIB metadata> */
 void p80211_totext_getmibattribute( struct catlistitem *metalist, UINT32 did, UINT8 *itembuf, char *textbuf );
