@@ -83,7 +83,6 @@
 /*================================================================*/
 /* Local Function Declarations */
 
-
 /*================================================================*/
 /* Function Definitions */
 
@@ -559,7 +558,7 @@ p80211meta_t *p80211_did2item( catlistitem_t *catlist, UINT32 did )
 UINT32 p80211item_getoffset( catlistitem_t *metalist, UINT32 did )
 {
 	UINT32			catgrp;
-	UINT32			offset;
+	UINT32			offset, len;
 	UINT32			tmpdid;
 	UINT32			item;
 	INT			i;
@@ -577,23 +576,14 @@ UINT32 p80211item_getoffset( catlistitem_t *metalist, UINT32 did )
 
 		for ( i = 1; i < item; i++ ) {
 			tmpdid = catgrp | P80211DID_MKITEM(i) | alist[i].did;
-			switch( P80211DID_TYPE(alist[i].did) ) {
-				case P80211_TYPE_OCTETSTR:
-				case P80211_TYPE_DISPLAYSTR:
-				case P80211_TYPE_INT:
-				case P80211_TYPE_BOUNDEDINT:
-				case P80211_TYPE_ENUMINT:
-				case P80211_TYPE_UNKDATA:
-					offset +=
-					p80211item_maxitemlen(metalist,tmpdid);
-					break;
-				default:
-					printf("Undefined data type for %s\n",
-						alist[i].name );
-					offset = 0xffffffff;
-					return offset;
-					break;
+			len = p80211item_maxdatalen(metalist, tmpdid);
+			if (len == 0xffffffffUL) {
+				printf("Undefined data type for %s\n",
+					alist[i].name );
+				offset = 0xffffffff;
+				return offset;
 			}
+			offset += len + sizeof(p80211item_t);
 		} /* for each arg meta data item up to current item */
 	} else {
 		offset = 0xffffffff;
@@ -603,6 +593,77 @@ UINT32 p80211item_getoffset( catlistitem_t *metalist, UINT32 did )
 }
 
 
+/*----------------------------------------------------------------
+* p80211item_gettype
+*
+* Returns the type of the item identified by a given DID.
+*
+* Arguments:
+*	meta	pointer to a metadata item
+*
+* Returns: 
+*	0		Unrecognized type in metadata.
+*	anything else	Success, return value is item type.
+----------------------------------------------------------------*/
+
+int p80211item_gettype(p80211meta_t *meta)
+{
+	int  type;
+
+	if (meta->totextptr == p80211_totext_octetstr)
+		type = P80211_TYPE_OCTETSTR;
+	else if (meta->totextptr == p80211_totext_displaystr)
+		type = P80211_TYPE_DISPLAYSTR;
+	else if (meta->totextptr == p80211_totext_boundedint)
+		type = P80211_TYPE_BOUNDEDINT;
+	else if (meta->totextptr == p80211_totext_int)
+		type = P80211_TYPE_INT;
+	else if (meta->totextptr == p80211_totext_enumint)
+		type = P80211_TYPE_ENUMINT;
+	else if (meta->totextptr == NULL ||
+		 meta->totextptr == p80211_totext_getmibattribute ||
+		 meta->totextptr == p80211_totext_setmibattribute)
+		type = P80211_TYPE_UNKDATA;
+	else if (meta->totextptr == p80211_totext_intarray)
+		type = P80211_TYPE_INTARRAY;
+	else if (meta->totextptr == p80211_totext_bitarray)
+		type = P80211_TYPE_BITARRAY;
+	else if (meta->totextptr == p80211_totext_macarray)
+		type = P80211_TYPE_MACARRAY;
+	else
+		type = 0;
+
+    return type;
+}
+
+
+/*----------------------------------------------------------------
+* p80211item_getaccess
+*
+* Returns the access type (read, write or read-write) of the item
+* identified by a given DID.
+*
+* Arguments:
+*	meta	pointer to a metadata item
+*
+* Returns: 
+*	0		Unrecognized access type in metadata.
+*	1		Read Only
+*	2		Write Only
+*	3		Read-Write
+----------------------------------------------------------------*/
+
+int p80211item_getaccess(p80211meta_t *meta)
+{
+	int  access_type;
+
+	access_type = 0;
+
+
+    return access_type;
+}
+
+
 /*----------------------------------------------------------------
 * p80211item_maxdatalen
 *
@@ -624,8 +685,11 @@ UINT32 p80211item_maxdatalen( catlistitem_t *metalist, UINT32 did )
 	p80211meta_t 	*meta;
 
 	if ( (meta = p80211_did2item(metalist, did)) != NULL ) {
-		switch( P80211DID_TYPE(did) ) {
+		switch( p80211item_gettype(meta) ) {
 			case P80211_TYPE_OCTETSTR:
+				/* add size of pstr length byte */
+				maxlen = meta->maxlen + sizeof(UINT8);
+				break;
 			case P80211_TYPE_DISPLAYSTR:
 				/* add size of pstr length byte */
 				maxlen = meta->maxlen + sizeof(UINT8);
@@ -634,10 +698,21 @@ UINT32 p80211item_maxdatalen( catlistitem_t *metalist, UINT32 did )
 			case P80211_TYPE_INT:
 			case P80211_TYPE_BOUNDEDINT:
 			case P80211_TYPE_ENUMINT:
+			case P80211_TYPE_BITARRAY:
 				/* all int types are 4 bytes */
 				maxlen = 4;
 				break;
 
+			case P80211_TYPE_INTARRAY:
+				/* int types are 4 bytes */
+				maxlen = meta->maxlen * 4;
+				break;
+
+			case P80211_TYPE_MACARRAY:
+				/* Addresses are 6 bytes, add a 4 byte count */
+				maxlen = (meta->maxlen * 6) + 4;
+				break;
+
 			case P80211_TYPE_UNKDATA:
 				maxlen = meta->maxlen;
 				break;
@@ -649,50 +724,14 @@ UINT32 p80211item_maxdatalen( catlistitem_t *metalist, UINT32 did )
 		maxlen = 0xffffffffUL;
 	}
 
-	return maxlen;
-}
-
-
-/*----------------------------------------------------------------
-* p80211item_maxitemlen
-*
-* Returns the total maximum data size of an item identified
-* by a given DID.  This is the length of the enitre triple.
-*
-* Arguments:
-*	metalist	pointer to a category metadata list
-*	did		A complete DID
-*
-* Returns: 
-*	0xffffffff	Bad DID or unrecognized type in metadata
-*	anything else	success, return value is maxdatalen
-----------------------------------------------------------------*/
-UINT32 p80211item_maxitemlen( catlistitem_t *metalist, UINT32 did )
-{
-	UINT32		maxlen;
-
-	switch( P80211DID_TYPE(did) ) {
-		case P80211_TYPE_OCTETSTR:
-		case P80211_TYPE_DISPLAYSTR:
-		case P80211_TYPE_INT:
-		case P80211_TYPE_BOUNDEDINT:
-		case P80211_TYPE_ENUMINT:
-		case P80211_TYPE_UNKDATA:
-			maxlen = p80211item_maxdatalen(metalist, did);
-			if (maxlen != 0xffffffffUL ) {
-				maxlen += sizeof(p80211item_t);
-			}
-			break;
-
-		default:
-			maxlen = 0xffffffffUL;
-			break;
-	}
+	/* pad for 32-bit aligmnent. */
+	if (maxlen != 0xffffffffUL)
+		if (maxlen % 4)
+			maxlen += (4 - (maxlen % 4));
 
 	return maxlen;
 }
 
-
 /*----------------------------------------------------------------
 * p80211_metaname2did
 *
@@ -730,7 +769,7 @@ UINT32 p80211_metaname2did(catlistitem_t *metalist, char *itemname)
 					[grp].itemlist[item].name) == 0 ) {
 
 					/* found it */
-					result = P80211DID_MKID( sec, grp, item, 0,0,0,0);
+					result = P80211DID_MKID( sec, grp, item, 0, 0, 0);
 					result |= metalist[sec].grplist
 					[grp].itemlist[item].did;
 				}
