@@ -56,14 +56,14 @@
 /* System Includes */
 
 #include <linux/config.h>
+#include <linux/version.h>
 #include <wlan/wlan_compat.h>
 
-#include <linux/version.h>
 #include <linux/module.h>
 #include <linux/kernel.h>
 #include <linux/sched.h>
 #include <linux/types.h>
-#include <linux/malloc.h>
+#include <linux/slab.h>
 #include <linux/netdevice.h>
 
 #include <pcmcia/config.h>
@@ -208,7 +208,9 @@ int wlanskel_mlmerequest(wlandevice_t *wlandev, p80211msg_t *msg)
 	ind->reasoncode.len = 4;
 	ind->reasoncode.data = 1;
 
+#ifdef CONFIG_NETLINK
 	p80211ind_mlme(wlandev, skb);
+#endif
 
 	return 0;
 }
